{% if mamedev_site %}
    <div class="rst-versions" data-toggle="rst-versions" role="note" aria-label="{{ _('Downloads') }}">
		<span class="rst-current-version" data-toggle="rst-current-version">
			Download documentation as <a href="https://docs.mamedev.org/_files/MAME.pdf">PDF</a> or <a href="https://docs.mamedev.org/_files/MAME.epub">EPUB</a>
		</span>
    </div>
{% endif %}
